import renpy
import pygame
import random
import os

spark_surfaces = dict()
im = renpy.display.im
gear_lock = im.Image('wires/gear-lock.png')
wires_selected = im.Image('wires/selected.png')
button_idle = im.Image('wires80/idle.png')
button_hovered = im.Image('wires80/hovered.png')
button_pressed = im.Image('wires80/pressed.png')
button_disabled = im.Image('wires80/disabled.png')
#button_locked = im.Image('wires80/locked.png')
wires_light = im.Image('wires/light.png')
nub_image = im.Image('wires80/nub.png')

def fuzz(src, radius, color):
    w, h = src.get_size()
    surf = renpy.display.pgrender.surface((w+(radius<<1), h+(radius<<1)),True)
    maxx, maxy = src.get_size()
    diam = (radius<<1)+1
    solid = renpy.display.pgrender.surface((diam,diam),True)
    #solid.fill(color)
    pygame.draw.circle(solid, color, (radius,radius), radius)
    y = 0
    while y<maxy:
        x = 0
        while x<maxx:
            sc = src.get_at((x,y))
            if sc[3]:
                surf.blit(solid,(x,y))
            x += 1
        y += 1
    return surf

def blurclone(surf):
    return fuzz(surf, 8, (0,0xd6,19,1))
    
class TextCache(dict):
    font = None
    def OpenFont(self):
        self.font = pygame.font.Font(os.path.join(renpy.store.config.searchpath[0],'wires80/ds-digib.ttf'),32)

    def __getitem__(self,args):
        if not self.has_key(args):
            if self.font is None:
                pygame.font.init()
                self.OpenFont()
            ret = self[args] = self.font.render(args[0], 1, args[1])
            self[args+('blur',)] = blurclone(ret)
        
        ret = dict.__getitem__(self,args)
        return ret

    def Free(self):
        # We use Free here rather than overriding clear because Ren'Py's
        # redefinition of dict() gets in the way of calling dict.clear()
        # on things.
        self.font = None
        self.clear()

text_cache = TextCache()

class ImCache(dict):
    def __getitem__(self,arg):
        if not self.has_key(arg):
            self[arg] = renpy.display.im.cache.get(arg)
        return dict.__getitem__(self,arg)
im_cache = ImCache()

def clean_cache(d):
    
    keys = d.keys()
    for k in keys:
        #if type(d[k]) != pygame.Surface:
        if type(d[k])==renpy.Render:
            d[k].kill()
        if nopickles.DEBUG:
            print "Debug: 77",d[k],type(d[k])
        del d[k]

def clean_caches():
    text_cache.Free() # should run this after every WiresMenu is done.
    clean_cache(text_cache)
    clean_cache(im_cache)
    spark_surfaces.clear()

def draw(dest, clip, what, xo, yo, screen, blit_flags=0):
    """
    This is the simple draw routine, which only works when alpha is 1.0
    and the matrices are None. If those aren't the case, draw_complex
    is used instead.

    `dest` - Either a destination surface, or a clipper.
    `clip` - If None, we should draw. Otherwise we should clip, and this is
    the rectangle to clip to.
    `what` - The Render or Surface we're drawing to.
    `xo` - The X offset.
    `yo` - The Y offset.
    `screen` - True if this is a blit to the screen, False otherwise.    
    """

    if not isinstance(what, renpy.display.render.Render):

        # Pixel-Aligned blit.
        if isinstance(xo, int) and isinstance(yo, int):
            if screen:
                what = renpy.display.im.rle_cache.get(id(what), what)

            if clip:
                w, h = what.get_size()
                dest.blits.append((xo, yo, xo + w, yo + h, clip, what, None))
            else:
                try:
                    renpy.display.render.blit_lock.acquire()
                    dest.blit(what, (xo, yo), None, blit_flags)
                finally:
                    renpy.display.render.blit_lock.release()
            
        # Subpixel blit.
        else:
            if clip:
                w, h = what.get_size()
                dest.blits.append((xo, yo, xo + w, yo + h, clip, what, None))
            else:            
                renpy.display.module.subpixel(what, dest, xo, yo)

        return

    # Deal with draw functions.
    if what.draw_func:

        xo = int(xo)
        yo = int(yo)

        if clip:
            dx0, dy0, dx1, dy1 = clip
            dw = dx1 - dx0
            dh = dy1 - dy0
        else:
            dw, dh = dest.get_size()
        
        if xo >= 0:
            newx = 0
            subx = xo
        else:
            newx = xo
            subx = 0

        if yo >= 0:
            newy = 0
            suby = yo
        else:
            newy = yo
            suby = 0

        if subx >= dw or suby >= dh:
            return

        # newx and newy are the offset of this render relative to the
        # subsurface. They can only be negative or 0, as otherwise we
        # would make a smaller subsurface.

        subw = min(dw - subx, what.width + newx) 
        subh = min(dh - suby, what.height + newy)

        if subw <= 0 or subh <= 0:
            return

        if clip:
            dest.forced.add((subx, suby, subx + subw, suby + subh, clip))
        else:
            newdest = dest.subsurface((subx, suby, subw, subh))
            what.draw_func(newdest, newx, newy)

        return

    # Deal with clipping, if necessary.
    if what.clipping:
        
        if clip:
            cx0, cy0, cx1, cy1 = clip

            cx0 = max(cx0, xo)
            cy0 = max(cy0, yo)
            cx1 = min(cx1, xo + what.width)
            cy1 = min(cy1, yo + what.height)

            if cx0 > cx1 or cy0 > cy1:
                return
            
            clip = (cx0, cy0, cx1, cy1)

        else:

            # After this code, x and y are the coordinates of the subsurface
            # relative to the destination. xo and yo are the offset of the
            # upper-left corner relative to the subsurface.
            
            if xo >= 0:
                x = xo
                xo = 0
            else:
                x = 0
                # xo = xo 

            if yo >= 0:
                y = yo
                yo = 0
            else:
                y = 0
                # yo = yo 

            dw, dh = dest.get_size()

            width = min(dw - x, what.width + xo)
            height = min(dh - y, what.height + yo)

            if width < 0 or height < 0:
                return
            
            dest = dest.subsurface((x, y, width, height))
        
    # Deal with alpha and transforms by passing them off to draw_transformed.
    if what.alpha != 1 or what.forward:
        for child, cxo, cyo, focus, main in what.visible_children:
            renpy.display.render.draw_transformed(dest, clip, child, xo + cxo, yo + cyo,
                             what.alpha, what.forward, what.reverse)
        return
        
    for child, cxo, cyo, focus, main in what.visible_children:
        draw(dest, clip, child, xo + cxo, yo + cyo, screen, blit_flags=what.blit_flags)

def render__init__(self, width, height, draw_func=None, layer_name=None, opaque=None, blit_flags=0):
        """/
        Creates a new render corresponding to the given widget with
        the specified width and height.

        If `layer_name` is given, then this render corresponds to a
        layer.
        """

        renpy.display.render.render_count += 1

        self.width = width
        self.height = height

        self.layer_name = layer_name

        # A list of (surface/render, xoffset, yoffset, focus, main) tuples, ordered from
        # back to front.
        self.children = [ ]

        # A list of Renders that are the parents of this Render. (We need
        # to kill these when this Render is redrawn.)
        self.parents = set()

        # A list of additional surfaces that we depend on. (Like children)
        self.depends_on_set = set()
        
        # A list of surfaces that depend on us.
        self.depends_on_us = set()

        # len(self.parents) + len(self.depends_on_us)
        self.refcount = 0
        
        # These are Matrix2D objects used to transform the children of
        # this render. If None, then no transformation is done. Otherwise,
        # they should be the inverse of each other. Forward is used to
        # project from render coordinates to child coordinates,
        # while reverse is used to project from render coordinates to
        # child coordinates.
        #
        # For performance reasons, these aren't used to transform the
        # x and y offsets found in self.children. Those offsets should
        # be of the (0, 0) point in the child coordinate space.
        self.forward = None
        self.reverse = None

        # This is used to adjust the alpha of children of this render.
        self.alpha = 1
        
        # A list of focus regions in this displayable.
        self.focuses = [ ]

        # Other renders that we should pass focus onto.
        self.pass_focuses = [ ]
        
        # The displayable(s) that this is a render of. (Set by render)
        self.render_of = [ ]

        # Is has this render been removed from the cache?
        self.cache_killed = False
        
        # Is this render dead?
        self.dead = False

        # If set, this is a function that's called to draw this render
        # instead of the default.
        self.draw_func = draw_func

        # Is this displayable opaque? (May be set on init, or later on
        # if we have opaque children.) This may be True, False, or None
        # to indicate we don't know yet.
        self.opaque = opaque

        # A list of our visible children. (That is, children above and
        # including our uppermost opaque child.) If nothing is opaque,
        # includes all children.
        self.visible_children = self.children
        
        # Should children be clipped to a rectangle?
        self.clipping = False

        # Caches of this render, rendered as a surface.
        self.surface = None
        self.alpha_surface = None
        self.blit_flags = blit_flags

from renpy.store import *

g_tint_values = (0.75,0.95,1.15,1.35,1.55)
g_wiresbuttons = dict()
class Spark(object):
        def __init__(self,x,y):
            self.x = x
            self.y = y
            if random.choice((True,False)):
                self.xvel = random.choice((-3,-4,-5,3,4,5))
                self.yvel = random.choice((0,0,0,-1,1))
            else:
                self.yvel = random.choice((-3,-4,-5,3,4,5))
                self.xvel = random.choice((0,0,0,-1,1))
            self.size = 4
            self.dying = False
            self.tintr = random.choice(g_tint_values)
            self.tintg = random.choice(g_tint_values)
            self.tintb = random.choice(g_tint_values)
class WireCell(object):
        def __init__(self,urdl,x,y,cellsize):
            object.__init__(self)
            self.u = urdl[0]
            self.d = urdl[1]
            self.l = urdl[2]
            self.r = urdl[3]
            self.angle = 0
            self.imgangle = 0
            if len(urdl)>4:
                self.zeroangle = urdl[4]
            else:
                self.zeroangle = 90
            self.lit = 0
            self.sparks = []
            self.x = x
            self.y = y
            self.xmin = x*cellsize
            self.xmax = self.xmin+cellsize-1
            self.ymin = y*cellsize
            self.ymax = self.ymin+cellsize-1
            if self.u and self.d and (not self.r) and (not self.l):
                self.imgangle = 90
        def Sparktick(self):
            kill = []
            for spark in self.sparks:
                if spark.dying or not self.lit:
                    spark.size -= 2
                    if spark.size<2:
                        kill.append(spark)
                elif spark.size<10:
                    spark.size += 2
                spark.x += spark.xvel
                spark.y += spark.yvel
                if spark.x<self.xmin:
                    kill.append(spark)
                    if self.l and self.x:
                        g_wiremap[self.y][self.x-1].sparks.append(spark)
                        spark.dying = True
                elif spark.x>self.xmax:
                    kill.append(spark)
                    if self.r and self.x<len(g_wiremap[self.y])-1:
                        g_wiremap[self.y][self.x+1].sparks.append(spark)
                        spark.dying = True
                elif spark.y<self.ymin:
                    kill.append(spark)
                    if self.u and self.y:
                        g_wiremap[self.y-1][self.x].sparks.append(spark)
                        spark.dying = True
                elif spark.y>self.ymax:
                    kill.append(spark)
                    if self.d and self.y<len(g_wiremap)-1:
                        g_wiremap[self.y+1][self.x].sparks.append(spark)
                        spark.dying = True
            if self.lit and not self.sparks:
                spark = Spark(self.xmin+32,self.ymin+32)
                if (spark.xvel>1 and not self.r):
                    pass
                elif spark.xvel<-1 and not self.l:
                    pass
                elif spark.yvel<-1 and not self.u:
                    pass
                elif spark.yvel>1 and not self.d:
                    pass
                else:
                    self.sparks.append(spark)
            newsparks = []
            for spark in self.sparks:
                if not spark in kill:
                    newsparks.append(spark)
            self.sparks = newsparks
        def __repr__(self):
            ret = "<WireCell "+str(self.x)+','+str(self.y)
            ret += ' '
            if self.u:
                ret += 'u'
            if self.r:
                ret += 'r'
            if self.d:
                ret += 'd'
            if self.l:
                ret += 'l'
            if self.angle:
                ret += '/'+str(self.angle)
            if self.lit:
                ret += ' !'
            ret += '>'
            return ret
        def Rotate(self):
            renpy.sound.play('wires/rotate_wire.ogg')
            self.angle = 10
        def RotationComplete(self):
            u = self.u
            self.u = self.l
            self.l = self.d
            self.d = self.r
            self.r = u
            self.imgangle += self.angle
            if self.imgangle>=self.zeroangle:
                self.imgangle = 0
            self.angle = 0
            g_wiresdisp.CheckLighting()
            renpy.restart_interaction()
        def Num(self):
            return self.u|(self.r<<1)|(self.d<<2)|(self.l<<3)
            # u = 1, r = 2, d = 4, l = 8
        def Surf(self):
            fn = ''
            if self.u:
                fn += 'u'
            if self.r:
                fn += 'r'
            if self.d:
                fn += 'd'
            if self.l:
                fn += 'l'
            if not fn:
                return None
            angle = self.angle+self.imgangle
            if (self.u and self.d) and (not self.r) and (not self.l):
                fn = 'rl'
            if self.lit:
                imgfn = 'wires/3-'+fn+'-lit.png'
            else:
                imgfn = 'wires/3-'+fn+'.png'
            if angle%90:
                return renpy.display.im.cache.get(im.Rotozoom('wires/3-'+fn+'.png',-angle,1))
            basesurf = renpy.display.im.cache.get(im.Image(imgfn))
            if angle:
                ret = pygame.transform.rotate(basesurf, -angle)
                return ret
            return basesurf
        
        def Img(self):
            raise Exception("Boom")
            fn = ''
            if self.u:
                fn += 'u'
            if self.r:
                fn += 'r'
            if self.d:
                fn += 'd'
            if self.l:
                fn += 'l'
            angle = self.angle+self.imgangle
            if (self.u and self.d) and (not self.r) and (not self.l):
                fn = 'rl'
            if self.lit:
                if angle:
                    return im.Rotozoom('wires/3-'+fn+'-lit.png',-angle,1)
                return im.Image('wires/3-'+fn+'-lit.png')
            elif angle:
                return im.Rotozoom('wires/3-'+fn+'.png',-angle,1)
            return im.Image('wires/3-'+fn+'.png')
class BlankWireCell(WireCell):
    def Img(self):
        return None
    def Rotate(self):
        pass
class LightCell(WireCell):
        def __init__(self,x,y):
            WireCell.__init__(self,(1,1,1,1,360),x,y,64)
            self.u = 1
            self.l = 1
            self.d = 1
            self.r = 1
        def Img(self):
            return im.Composite((64,64),
                (0,0),'wires/3-urdl-lit.png',
                #(4,4),'wires/frame.png',
                (4,4),'wires/light.png')
        def Surf(self):
            ret = WireCell.Surf(self).convert_alpha()
            ret.blit(renpy.display.im.cache.get(wires_light),(4,4))
            return ret
class ButtonMasterCell(WireCell):
        def __repr__(self):
            return "<Button "+self.title+" "+WireCell.__repr__(self)+">"
        def __init__(self,x,y,title,buttonwidth=3,locked=False,clickedsound='wires/button.ogg'):
            WireCell.__init__(self,(0,0,0,1),x,y,64)
            self.title = title
            self.buttonwidth = buttonwidth
            self.locked = locked
            self.l = 1
            self.r = self.u = self.d = 0
            self.clickedsound = clickedsound
        def Img(self):
            g_wiresbuttons[self.title] = self.lit
            return None
            #if self.lit:
            #    return im.Image('wires/button.png')#im.Alpha('wires/button.png',0.9)
            #return im.Image('wires/button-inact.png')#Alpha('wires/button-inact.png',0.5)
        def Rotate(self):
            if self.lit:
                if self.clickedsound:
                    renpy.sound.play(self.clickedsound)
                #renpy.sound.play('wires/button.ogg')
                return self.title
class ButtonSlaveCell(WireCell):
        def __init__(self,x,y,title):
            WireCell.__init__(self,(0,0,0,1),x,y,64)
            self.title = title
            self.l = self.r = self.u = self.d
        def Img(self):
            return None
        def Rotate(self):
            pos = g_wiresdisp.buttons[self.title]
            if g_wiremap[pos[1]][pos[0]].lit:
                if g_wiremap[pos[1]][pos[0]].clickedsound:
                    renpy.sound.play(g_wiremap[pos[1]][pos[0]].clickedsound)
                #renpy.sound.play('wires/button.ogg')
                return self.title

def init_spark_surfaces():
    for r in g_tint_values:
        for g in g_tint_values:
            for b in g_tint_values:
                for s in (2,4,6,8,10):
                    img = renpy.display.im.cache.get(im.Scale(
                        im.MatrixColor('wires/flare.png',im.matrix.tint(r,g,b)),
                            s,s))
                    # For old versions of Ren'Py
                    #def additive_draw(dsurf,dx,dy,surf=img):
                    #    if renpy.display.scale.factor!=1.0:
                    #        dsurf.surface.blit(surf.surface,(dx,dy),None,1)
                    #        return
                    #    dsurf.blit(surf,(dx,dy),None,1)
                    #flarerender = renpy.Render(s,s,draw_func=additive_draw)
                    spark_surfaces[(r,g,b,s)] = img
def deinit_spark_surfaces():
    for v in spark_surfaces.values():
        v.kill()
    spark_surfaces = dict()

lock_image = im.Image('wires80/lock.png')

class WiresDisplayable(renpy.Displayable):
        focusable = True
        buttoff = 17
        no_randoms = False
        def __init__(self,cellsize=64,width=14,height=10,lightpos=(7,1),buttons=(),grid=(),offset=-4):
            renpy.Displayable.__init__(self,xmaximum=cellsize*width,
                ymaximum=cellsize*height)
            self.cellsize = cellsize
            cells = []
            y = 0
            while len(cells)<height:
                row = []
                x = 0
                while len(row)<width:
                    if y<len(grid) and x<len(grid[y]):
                        urdl = self.wire_for(grid[y][x])
                    else:
                        urdl = random.choice((
                            (1,0,0,0),
                            (1,1,1,0),
                            (1,1,0,0,360),
                            (1,1,1,1,360),
                            ))
                    x += 1
                    row.append(WireCell(urdl,len(row),len(cells),cellsize))
                y += 1
                cells.append(row)
            self.buttons = dict()
            for button in buttons:
                buttlocked = False
                if len(button)>3:
                    buttlen = button[3]
                    if len(button)>4:
                        buttlocked = True
                else:
                    buttlen = 3
                self.buttons[button[2]] = (button[0],button[1],buttlen,buttlocked)
                cells[button[1]][button[0]] = ButtonMasterCell(*button)
                if len(button)>3:
                    buttslavex = button[0]+1
                    endbuttslave = button[0]+button[3]
                    while buttslavex<endbuttslave:
                        cells[button[1]][buttslavex] = ButtonSlaveCell(buttslavex,button[1],button[2])
                        buttslavex += 1
                else:
                    cells[button[1]][button[0]+1] = ButtonSlaveCell(button[0]+1,button[1],button[2])
                    cells[button[1]][button[0]+2] = ButtonSlaveCell(button[0]+2,button[1],button[2])
            self.pixw = cellsize*width
            self.pixh = cellsize*height
            self.width = width
            self.height = height
            cells[lightpos[1]][lightpos[0]] = LightCell(lightpos[0],lightpos[1])
            self.cells = cells
            self.lightpos = lightpos
            self.CheckLighting()
            self.offset = offset
            self.cursor = lightpos
            self.mouse = False
        def CheckLighting(self):
            for row in self.cells:
                for cell in row:
                    cell.lit = 0
            tocheck = [self.lightpos,]
            x = 0
            while x<len(tocheck):
                pos = tocheck[x]
                if self.cells[pos[1]][pos[0]].d and pos[1]<len(self.cells)-1 and self.cells[pos[1]+1][pos[0]].u:
                    if not self.cells[pos[1]+1][pos[0]].lit:
                        self.cells[pos[1]+1][pos[0]].lit = 1
                        tocheck.append((pos[0],pos[1]+1))
                if self.cells[pos[1]][pos[0]].u and pos[1] and self.cells[pos[1]-1][pos[0]].d and not self.cells[pos[1]-1][pos[0]].lit:
                    self.cells[pos[1]-1][pos[0]].lit = 1
                    tocheck.append((pos[0],pos[1]-1))
                if self.cells[pos[1]][pos[0]].r:
                    if pos[0]<len(self.cells[pos[1]])-1 and self.cells[pos[1]][pos[0]+1].l and not self.cells[pos[1]][pos[0]+1].lit:
                        self.cells[pos[1]][pos[0]+1].lit = 1
                        if type(self.cells[pos[1]][pos[0]+1])==ButtonMasterCell:
                            if self.cells[pos[1]][pos[0]+1].locked:
                                self.cells[pos[1]][pos[0]+1].lit=False
                            else:
                                renpy.restart_interaction()
                        else:
                            tocheck.append((pos[0]+1,pos[1]))
                if self.cells[pos[1]][pos[0]].l and pos[0] and self.cells[pos[1]][pos[0]-1].r and not self.cells[pos[1]][pos[0]-1].lit:
                    self.cells[pos[1]][pos[0]-1].lit = 1
                    tocheck.append((pos[0]-1,pos[1]))
                x += 1
        def wire_for(disp, c):
            if c == '/':
                return random.choice(((1,1,0,0,360),(0,0,1,1,360)))
            elif c == '|':
                return (1,1,0,0,360)
            elif c == '-':
                return (0,0,1,1,360)
            elif c == 'L':
                return random.choice(((0,1,0,1),(1,0,1,0),(1,0,0,1),(0,1,1,0)))
            elif c == 'T':
                return random.choice(((1,1,1,0),(1,1,0,1),(1,0,1,1),(0,1,1,1)))
            elif c == '+':
                return (1,1,1,1,360)
            elif c == '.':
                return random.choice(((1,0,0,0),(0,1,0,0),(0,0,1,0),(0,0,0,1)))
            elif c == '0':
                return (0,0,0,0)
            # *, B, and b are reserved for light and buttons respectively.
            elif c ==u'├':
                return (1,1,0,1)
            elif c ==u'┴':
                return (1,0,1,1)
            elif c ==u'┬':
                return (0,1,1,1)
            elif c ==u'┘':
                return (1,0,1,0)
            elif c ==u'└':
                return (1,0,0,1)
            elif c ==u'┐':
                return (0,1,1,0)
            elif c == u'┌':
                return (0,1,0,1)
            else:
                if WiresDisplayable.no_randoms:
                    return (0,0,0,0)
                return random.choice((
                    (1,0,0,0),
                    (1,1,1,0),
                    (1,1,0,0,360),
                    (1,1,1,1,360),
                    ))
        def render(disp, w, h, st, at):
            if not spark_surfaces:
                init_spark_surfaces()
            ret = renpy.Render(disp.pixw,disp.pixh)

            # This is the transparent version, which gives the impression the
            # sparks crawl among the wires.
            #puzzlesurf = renpy.display.pgrender.surface((disp.pixw, disp.pixh), True)

            # Opaque version
            puzzlesurf = renpy.display.pgrender.surface((disp.pixw, disp.pixh), False)

            puzzlesurf.blit(renpy.display.im.cache.get(im.Crop('wires80/bg.jpg',253,53,disp.pixw, disp.pixh)),(0,0))
            y = 3
            for row in disp.cells:
                x = 3
                for cell in row:
                    surf = cell.Surf()
                    if surf is not None:
                        xdelt = disp.offset
                        ydelt = disp.offset
                        if cell.angle:
                            cell.angle += 10
                            if cell.angle>=90:
                                cell.RotationComplete()
                                surf = cell.Surf()
                        if surf.get_size()!=(72,72):
                            cellw,cellh = surf.get_size()
                            xdelt = (disp.cellsize-cellw)/2
                            ydelt = (disp.cellsize-cellh)/2
                        puzzlesurf.blit(surf, (x+xdelt,y+ydelt))
                        cell.Sparktick()
                        for spark in cell.sparks:
                            puzzlesurf.blit(spark_surfaces[(spark.tintr,spark.tintg,spark.tintb,spark.size)],(spark.x,spark.y),None, pygame.BLEND_ADD)
                    x += disp.cellsize
                y += disp.cellsize
            if disp.mouse:
                x, y = renpy.display.draw.get_mouse_pos()#pygame.mouse.get_pos()
                x -= 253
                y -= 53
                #x -= 62
                #y -= 62
                x -= x%64
                y -= y%64
            else:
                x = disp.cursor[0]*64
                y = disp.cursor[1]*64
            puzzlesurf.blit(renpy.display.im.cache.get(im.Image('wires80/grid.png')),(0,0))
            delayed_blits = []
            for button in disp.buttons:
                bx,by,bs = disp.buttons[button][:3]
                bxpo = bxp = bx*64
                bypo = byp = by*64
                bsp = bs*64
                bxp -= 4
                byp -= 5
                if disp.cells[by][bx].lit:
                    if x>=bxp and x<bxp+bsp and y>=byp and y<byp+64:
                        disp.BlitButton(puzzlesurf, button_hovered, bxp, byp, bsp)
                        blur = text_cache[button,(0,0xd6,0x19),bsp-20,64,'blur']
                        textw, texth = blur.get_size()
                        delayed_blits.append((blur,(bxpo+disp.buttoff+((bsp-disp.buttoff)>>1)-(int(textw)>>1),bypo+32-(int(texth)>>1))))
                    else:
                        disp.BlitButton(puzzlesurf, button_idle, bxp, byp, bsp,lit=True)
                    text = text_cache[button,(0,0xd6,0x19),bsp-20,64]
                    # Hovered: 00d619, Idle: 00d619, Inactive 5a5a5a
                elif disp.cells[by][bx].locked:
                    #disp.BlitButton(puzzlesurf, button_locked, bxp, byp, bsp,locked=True)
                    disp.BlitButton(puzzlesurf, button_disabled, bxp, byp, bsp,locked=True)
                    text = text_cache[button,#(0x5a,0x5a,0x5a),
                        (0x80,0x80,0x80),
                        bsp-20,64]
                    locksurf = renpy.display.im.cache.get(lock_image)
                    puzzlesurf.blit(locksurf, (bxp-64, byp))
                else:
                    disp.BlitButton(puzzlesurf, button_disabled, bxp, byp, bsp)
                    text = text_cache[button,#(0x5a,0x5a,0x5a),
                        (0xc0,0xc0,0xc0),
                        bsp-20,64]
                textw, texth = text.get_size()
                puzzlesurf.blit(text, (bxpo+disp.buttoff+((bsp-disp.buttoff)>>1)-(int(textw)>>1),bypo+32-(int(texth)>>1)))
                #if disp.cells[by][bx].locked:
                #    puzzlesurf.blit(renpy.display.im.cache.get(gear_lock),
                #        (bxp+(bsp>>1)-48,byp-16))
                #    text = text_cache['Locked',(0x80,0x80,0x80),bsp-20,64]
                #    textsize = text.get_size()
                #    puzzlesurf.blit(text,(bxp-6+((bsp/2)+8)-(textsize[0]/2),byp+30-(textsize[1]/2)))
            if y<disp.pixh and x<disp.pixw and x>=0 and y>=0:
                cursor = renpy.display.im.cache.get(im.Image('wires/cursor.png'))
                puzzlesurf.blit(cursor,(x,y))
            renpy.redraw(g_wiresdisp,0.05)
            for blit in delayed_blits:
                puzzlesurf.blit(*blit)
            ret.blit(puzzlesurf,(0,0))
            delayed_blits = None
            return ret
        def BlitButton(self, puzzlesurf, butt, bxp, byp, bsp, lit=False, locked=False):
            buttsurf = renpy.display.im.cache.get(butt)
            puzzlesurf.blit(renpy.display.im.cache.get(nub_image),(bxp,byp+24))
            bw, bh = buttsurf.get_size()
            puzzlesurf.blit(buttsurf,(bxp+self.buttoff,byp),(0,0,32,bh))

            bonusw = bw-192
            targw = bsp+bonusw-self.buttoff
            bmid = pygame.transform.scale(buttsurf.subsurface(32,0,bw-64,bh),(targw-64,bh))
            # For some reason, I get a visible seam on normal buttons
            # if I use bw-32 but not bw-33.  This might be related to Ren'Py's
            # padding.
            puzzlesurf.blit(buttsurf.subsurface(bw-33,0,33,bh),(bxp+targw-self.buttoff,byp))
            puzzlesurf.blit(bmid, (bxp+32+self.buttoff, byp))
            return

        def Clicked(ui,x,y):
            if x<0 or y<0:
                return
            x = int(x/ui.cellsize)
            y = int(y/ui.cellsize)
            if y<len(ui.cells) and x<len(ui.cells[y]):
                if ui.cells[y][x].angle:
                    ui.cells[y][x].angle = 90
                    ui.cells[y][x].RotationComplete()
                return ui.cells[y][x].Rotate()
        def cursor_cell(ui):
            x, y = pygame.mouse.get_pos()
            x -= 62
            y -= 62
            x -= x%64
            y -= y%64
            x /= 64
            y /= 64
            return (x,y)
        def CursorUp(ui):
            if ui.mouse:
                x,y = ui.cursor_cell()
            else:
                x,y = ui.cursor
            if y>0:
                ui.cursor = (x,y-1)
            if y==ui.height:
                renpy.display.focus.change_focus(None)
            raise renpy.display.core.IgnoreEvent()
        def CursorDown(ui):
            if ui.mouse:
                x,y = ui.cursor_cell()
            else:
                x,y = ui.cursor
            if y<ui.height:
                ui.cursor = (x,y+1)
                if y+1<ui.height:
                    raise renpy.display.core.IgnoreEvent()
        def CursorLeft(ui):
            if ui.mouse:
                x,y=ui.cursor_cell()
            else:
                x,y = ui.cursor
            if x>0:
                ui.cursor = (x-1,y)
                ui.mouse = False
            raise renpy.display.core.IgnoreEvent()
        def CursorRight(ui):
            if ui.mouse:
                x,y = ui.cursor_cell()
            else:
                x,y = ui.cursor
            if x<ui.width-1:
                ui.mouse = False
                ui.cursor = (x+1,y)
            raise renpy.display.core.IgnoreEvent()
        def RotateCell(ui):
            if ui.mouse:
                x,y = pygame.mouse.get_pos()
                ui.cursor = ui.cursor_cell()
                return ui.Clicked(x-62,y-62)
            else:
                ui.mouse = False
                return ui.Clicked((ui.cursor[0]*64),(ui.cursor[1]*64))

        def event(ui, ev, x, y, st):
            if ev.type==pygame.MOUSEMOTION:
                ui.mouse = True
            elif ev.type==pygame.MOUSEBUTTONUP:
                if ev.button == 1:
                    return ui.Clicked(x,y)
            elif ev.type==pygame.KEYDOWN:
                if ev.key==pygame.K_DOWN:
                    ui.CursorDown()
                elif ev.key==pygame.K_UP:
                    ui.CursorUp()
                elif ev.key==pygame.K_LEFT:
                    ui.CursorLeft()
                elif ev.key==pygame.K_RIGHT:
                    ui.CursorRight()
                elif ev.key==pygame.K_SPACE or ev.key==pygame.K_KP_ENTER or ev.key==pygame.K_RETURN:
                    return ui.RotateCell()
                else:
                    return
                ui.mouse = False
            elif ev.type==renpy.display.core.JOYEVENT:
                joykey = unroller.inverse_joymap.get(ev.press,None)
                if joykey=='joy_up':
                    ui.CursorUp()
                elif joykey=='joy_down':
                    ui.CursorDown()
                elif joykey=='joy_right':
                    ui.CursorRight()
                elif joykey=='joy_left':
                    ui.CursorLeft()
                elif joykey=='joy_dismiss':
                    return ui.RotateCell()
                ui.mouse = False

class WiresGalleryDisplayable(WiresDisplayable):
        def render(disp, w, h, st, at):
            if not spark_surfaces:
                init_spark_surfaces()
            r = renpy.Render(disp.pixw,disp.pixh)
            y = 0
            #framerender = im.Image('wires/frame.png').render(64,64,0,0)
            for row in disp.cells:
                x = 0
                for cell in row:
                    img = cell.Img()
                    if img is not None:
                        xdelt = disp.offset
                        ydelt = disp.offset
                        surf = renpy.display.im.cache.get(cell.Img())
                        if cell.angle:
                            cell.angle += 10
                            if cell.angle>=90:
                                cell.RotationComplete()
                                surf = renpy.display.im.cache.get(cell.Img())
                            if surf.get_size()!=(64,64):
                                cellw,cellh = surf.get_size()
                                xdelt = (disp.cellsize-cellw)/2
                                ydelt = (disp.cellsize-cellh)/2
                        puzzlesurf.blit(surf, (x+xdelt,y+ydelt))
                        cell.Sparktick()
                        for spark in cell.sparks:
                            puzzlesurf.blit(spark_surfaces[(spark.tintr,spark.tintg,spark.tintb,spark.size)],(spark.x,spark.y))
                    x += disp.cellsize
                y += disp.cellsize
            if disp.mouse:
                x, y = pygame.mouse.get_pos()
                x -= 62
                y -= 62
                x -= x%64
                y -= y%64
            else:
                x = disp.cursor[0]*64
                y = disp.cursor[1]*64
            puzzlesurf.blit(renpy.display.im.cache.get(im.Image('wires/grid.png')),(0,0))
            buttoni = 0
            for button in disp.buttons:
                bx,by,bs = button[:3]
                bxp = bx*64
                byp = by*64
                bsp = bs*64
                if disp.cells[by][bx].lit:
                    text = text_cache[str(buttoni)]
                else:
                    text = text_cache[str(buttoni)]
                textsize = text.get_size()
                textx = ((bsp/2)+8)-(textsize[0]/2)
                puzzlesurf.blit(text,(bxp+14,byp+4))
                if disp.cells[by][bx].lit:
                    if x>=bxp and x<bxp+bsp and y>=byp and y<byp+64:
                        puzzlesurf.blit(im_cache[im.Image('ui/gallerybutton-active-hover.png')],(bxp,byp))
                    else:
                        puzzlesurf.blit(im_cache[im.Image('ui/gallerybutton-active.png')],(bxp,byp))
                else:
                        puzzlesurf.blit(im_cache[im.Image('ui/gallerybutton-inact.png')],(bxp,byp))
                if disp.cells[by][bx].locked:
                    puzzlesurf.blit(im_cache[im.Image('wires/gear-lock.png')],(bxp+(bsp>>1)-48,byp-16))
                    text = text_cache['Locked','#fff',bsp-20,64]
                    textsize = text.get_size()
                    puzzlesurf.blit(text,(bxp-6+((bsp/2)+8)-(textsize[0]/2),byp+30-(textsize[1]/2)))
                buttoni += 1
            if y<disp.pixh and x<disp.pixw and x>=0 and y>=0:
                cursor = renpy.display.im.cache.get(im.Image('wires/cursor.png'))
                puzzlesurf.blit(cursor,(x,y))
            renpy.redraw(g_wiresdisp,0.05)
            ret.blit(puzzlesurf,(0,0))
            return ret
        def CheckLighting(self):
            for row in self.cells:
                for cell in row:
                    cell.lit = 0
            tocheck = [self.lightpos,]
            x = 0
            while x<len(tocheck):
                pos = tocheck[x]
                if self.cells[pos[1]][pos[0]].d and pos[1]<len(self.cells)-1 and self.cells[pos[1]+1][pos[0]].u:
                    if not self.cells[pos[1]+1][pos[0]].lit:
                        self.cells[pos[1]+1][pos[0]].lit = 1
                        tocheck.append((pos[0],pos[1]+1))
                if self.cells[pos[1]][pos[0]].u and pos[1] and self.cells[pos[1]-1][pos[0]].d and not self.cells[pos[1]-1][pos[0]].lit:
                    self.cells[pos[1]-1][pos[0]].lit = 1
                    tocheck.append((pos[0],pos[1]-1))
                if self.cells[pos[1]][pos[0]].r:
                    if pos[0]<len(self.cells[pos[1]])-1 and self.cells[pos[1]][pos[0]+1].l and not self.cells[pos[1]][pos[0]+1].lit:
                        self.cells[pos[1]][pos[0]+1].lit = 1
                        if type(self.cells[pos[1]][pos[0]+1])==ButtonMasterCell:
                            if self.cells[pos[1]][pos[0]+1].locked:
                                self.cells[pos[1]][pos[0]+1].lit=False
                            else:
                                tocheck.append((pos[0]+1,pos[1]))
                                renpy.restart_interaction()
                        else:
                            tocheck.append((pos[0]+1,pos[1]))
                if self.cells[pos[1]][pos[0]].l and pos[0] and self.cells[pos[1]][pos[0]-1].r and not self.cells[pos[1]][pos[0]-1].lit:
                    self.cells[pos[1]][pos[0]-1].lit = 1
                    tocheck.append((pos[0]-1,pos[1]))
                x += 1
        def printcells(self):
            for row in self.cells:
                for cell in row:
                    if cell.lit:
                        print '!',
                    else:
                        print '.',
                    if cell.u:
                        if cell.d:
                            if cell.r:
                                if cell.l:
                                    print u'┼',
                                else:
                                    print u'├',
                            else:
                                if cell.l:
                                    print u'┥',
                                else:
                                    print u'│',
                        else:
                            if cell.l:
                                if cell.r:
                                    print u'┴',
                                else:
                                    print u'┙',
                            else:
                                if cell.r:
                                    print u'└',
                                else:
                                    print u'╵',
                    elif cell.d: # No up
                        if cell.l:
                            if cell.r:
                                print u'┬',
                            else:
                                print u'┐',
                        elif cell.r:
                            print u'┌',
                        else:
                            print u'╷',
                    elif cell.l:
                        if cell.r:
                            print u'─',
                        else:
                            print u'╸',
                    elif cell.r:
                        print u'╶',
                    else:
                        print 'o',
                print ''
        def __init__(self,cellsize=64,width=14,height=10,lightpos=(7,1),buttons=(),grid=(),offset=-4):
            renpy.Displayable.__init__(self,xmaximum=cellsize*width,
                ymaximum=cellsize*height)
            self.cellsize = cellsize
            cells = []
            y = 0
            while len(cells)<height:
                row = []
                x = 0
                while len(row)<width:
                    if y<len(grid) and x<len(grid[y]):
                        urdl = self.wire_for(grid[y][x])
                    elif WiresDisplayable.no_randoms:
                        urdl = (0,0,0,0)
                    else:
                        urdl = random.choice((
                            (1,0,0,0),
                            (1,1,1,0),
                            (1,1,0,0,360),
                            (1,1,1,1,360),
                            ))
                    x += 1
                    if urdl!=(0,0,0,0):
                        row.append(WireCell(urdl,len(row),len(cells),cellsize))
                    else:
                        row.append(BlankWireCell(urdl,len(row),len(cells),cellsize))
                y += 1
                cells.append(row)
            self.buttons = []
            for button in buttons:
                buttlocked = False
                if len(button)>3:
                    buttlen = button[3]
                    if len(button)>4:
                        buttlocked = True
                else:
                    buttlen = 3
                self.buttons.append((button[0],button[1],buttlen,buttlocked))
                cells[button[1]][button[0]] = ButtonMasterCell(*button,**dict(clickedsound=None))
                if len(button)>3:
                    buttslavex = button[0]+1
                    endbuttslave = button[0]+button[3]
                    while buttslavex<endbuttslave:
                        cells[button[1]][buttslavex] = ButtonSlaveCell(buttslavex,button[1],button[2])
                        cells[button[1]][buttslavex].title = len(self.buttons)-1
                        buttslavex += 1
                else:
                    cells[button[1]][button[0]+1] = ButtonSlaveCell(button[0]+1,button[1],button[2])
                    cells[button[1]][button[0]+2] = ButtonSlaveCell(button[0]+2,button[1],button[2])
            self.pixw = cellsize*width
            self.pixh = cellsize*height
            self.width = width
            self.height = height
            cells[lightpos[1]][lightpos[0]] = LightCell(lightpos[0],lightpos[1])
            self.cells = cells
            self.lightpos = lightpos
            self.CheckLighting()
            self.offset = offset
            self.cursor = lightpos
            self.mouse = False
